"use strict";
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = void 0;
var yargs_parser_1 = __importDefault(require("yargs-parser"));
var forEach_1 = __importDefault(require("lodash/forEach"));
var debug_1 = require("../debug");
/**
 * list missing args
 *
 * @param {Object} appCmd - command definition
 * @param {Array} positionalArgs from commandline
 * @returns {Array[]} missing args
 */
function findMissingArgs(appCmd, positionalArgs) {
    var missing = [];
    forEach_1.default(appCmd.args, function (arg, idx) {
        if (!positionalArgs[idx]) {
            if (!arg.isOptional) {
                missing.push(arg);
            }
        }
    });
    return missing;
}
/**
 * clean and format options specified in the command
 *
 * @param {Object} appCmd - command definition
 * @param {Object} argv - options parsed from commandline
 * @returns {Object} argv and unknown options
 */
function processOptions(appCmd, argv) {
    // console.log("processOptions", JSON.stringify({ appCmd, argv }, null, 2));
    var newArgv = {};
    var unknownOpts = [];
    forEach_1.default(argv, function (val, key) {
        var _a;
        var _b, _c;
        // don't add aliases
        if (!((_b = appCmd.aliasLookup) === null || _b === void 0 ? void 0 : _b[key])) {
            // ensure options are arrays unless boolean
            if (val !== true && val !== false) {
                val = Array.isArray(val) ? val : [val];
            }
            // classify unknown options
            if (((_c = appCmd.optsLookup) === null || _c === void 0 ? void 0 : _c[key]) === undefined) {
                unknownOpts.push((_a = {}, _a[key] = val, _a));
            }
            else {
                newArgv[key] = val;
            }
        }
    });
    return { unknownOpts: unknownOpts, args: newArgv };
    // return { unknownOpts, options: newArgv };
}
/**
 *
 * @param {Number} argCount - the number of args on commandline
 * @param {Boolean} nextArg - true if the last arg on commandline is complete
 * @param {Array} argsDef - the command arguments definition
 * @returns {String} the current arg
 */
function getCurrentArg(argCount, nextArg, argsDef) {
    var current;
    if (argsDef && argsDef.length) {
        var lastIdx = argsDef.length - 1;
        var idx = argCount;
        if (idx && !nextArg) {
            idx--;
        }
        if (idx < argsDef.length) {
            current = argsDef[idx].name;
        }
        else if (argsDef[lastIdx].isMulti) {
            current = argsDef[lastIdx].name;
        }
    }
    return current;
}
/**
 * @typedef {Object} Args
 * @property {Array} missing - required arguments not provided
 * @property {Array} unknownArgs - arguments which do not exist in the command definition
 * @property {Array} unknownOpts - options which do not exist in the command definition
 * @property {Object} args - all valid arguments and options from commandline
 */
/**
 * parse command args and options
 *
 * @param {Object} appCmd - command definition
 * @param {String} cmdArgs - arguments and options from commandline
 * @param {String} cmdLine - commandLine string
 * @returns {Args} the parsed commandline
 */
function parseArgs(appCmd, cmdArgs, cmdLine) {
    var _a;
    var _b;
    if (cmdLine === void 0) { cmdLine = ""; }
    var parsed = yargs_parser_1.default.detailed(cmdArgs, appCmd.argsParserOpts);
    if (parsed.error) {
        throw parsed.error;
    }
    // eslint-disable-next-line prefer-const
    // let { _: positionalArgs, ...args } = parsed.argv;
    // eslint-disable-next-line prefer-const
    var _c = parsed.argv, positionalArgs = _c._, script = _c.$0, args = __rest(_c, ["_", "$0"]);
    // if positional args have no value they need to appear in missing
    var missing = findMissingArgs(appCmd, positionalArgs);
    var unknownOpts;
    // eslint-disable-next-line prefer-const
    (_a = processOptions(appCmd, args), args = _a.args, unknownOpts = _a.unknownOpts);
    // eslint-disable-next-line prefer-const
    // let { options, unknownOpts } = processOptions(appCmd, args);
    var unknownArgs = [];
    var lastIdx = appCmd.args ? appCmd.args.length - 1 : -1;
    var lastArg = lastIdx >= 0 && ((_b = appCmd.args) === null || _b === void 0 ? void 0 : _b[lastIdx])
        ? appCmd.args[lastIdx]
        : { isMulti: false, name: "" };
    // assign positional values
    forEach_1.default(positionalArgs, function (value, idx) {
        if (appCmd.args && appCmd.args[idx]) {
            if (appCmd.args[idx].isMulti) {
                args[appCmd.args[idx].name] = [value];
            }
            else {
                args[appCmd.args[idx].name] = value;
            }
        }
        else if (lastArg === null || lastArg === void 0 ? void 0 : lastArg.isMulti) {
            // if (Array.isArray(args[lastArg.name])) {
            args[lastArg.name].push(value);
        }
        else {
            //     args[lastArg.name] = value;
            //   }
            // } else {
            unknownArgs.push(value);
        }
    });
    var argCount = positionalArgs.length;
    var nextArg = /\s$/.test(cmdLine);
    var current = getCurrentArg(argCount, nextArg, appCmd.args);
    var result = { args: args, missing: missing, current: current };
    // const result: ParsedArgs = { args, opts: options, missing, current };
    if (unknownArgs.length) {
        result.unknownArgs = unknownArgs;
    }
    if (unknownOpts.length) {
        result.unknownOpts = unknownOpts;
    }
    return result;
}
/**
 *
 * @param {Object} arg0 -
 * @param {Object} arg0.appCmd - command definition
 * @param {String} arg0.cmdArgs - arguments and options input
 * @returns {Args} parsed args with validation
 */
function parse(arg0) {
    var appCmd = arg0.appCmd, cmdArgs = arg0.cmdArgs, cmdLine = arg0.cmdLine;
    debug_1.debug({ appCmd: appCmd, cmdArgs: cmdArgs });
    var parsedArgs = parseArgs(appCmd, cmdArgs, cmdLine);
    debug_1.debug({ parsedArgs: parsedArgs });
    return parsedArgs;
}
exports.parse = parse;
