"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Makitso = void 0;
var merge_1 = __importDefault(require("lodash/merge"));
var flatten_1 = __importDefault(require("lodash/flatten"));
var isString_1 = __importDefault(require("lodash/isString"));
var makitso_prompt_1 = require("makitso-prompt");
var command_1 = require("./command");
var context_1 = require("./context");
var debug_1 = require("./debug");
/**
 *
 * @param {String} arg - name and description
 * @returns {Object} arg
 */
function parseArgument(arg) {
    var args = arg.split(/\s+/);
    var name = args[0];
    var description = args.slice(1);
    var isMulti;
    var isOptional;
    if (/^\[/.test(name)) {
        isOptional = true;
        name = name.replace(/[[\]]/g, "");
    }
    if (/\.\.\.$/.test(name)) {
        isMulti = true;
        name = name.replace(/\.\.\.$/, "");
    }
    return {
        name: name,
        description: description.join(" "),
        isMulti: Boolean(isMulti),
        isOptional: Boolean(isOptional),
        string: arg,
    };
}
/**
 *
 * @param {String} opt - flags and description
 * @returns {Object} opt
 */
function parseOption(opt) {
    var optBits = opt.split(/\s+/);
    debug_1.debug({ optBits: optBits });
    var initial = {
        name: "",
        alias: "",
        description: "",
        parseOpt: { boolean: false },
    };
    var option = optBits.reduce(function (result, val) {
        if (/^--\w/.test(val)) {
            result.name = val.replace(/^--/, "");
        }
        else if (/^-\w/.test(val)) {
            result.alias = val.replace(/^-/, "");
        }
        else if (/^\{\w+}$/.test(val)) {
            var parseOpt = val.replace(/^\{/, "").replace(/}$/, "");
            if (parseOpt === "bool" || parseOpt === "boolean") {
                result.parseOpt.boolean = true;
            }
        }
        else {
            result.description = result.description + " " + val;
        }
        return result;
    }, initial);
    return option;
}
function optionsToArgsParserOpts(options) {
    return options.reduce(function (result, opt) {
        if (result === void 0) { result = {}; }
        result[opt.name] = opt.alias;
        return result;
    }, {});
}
function parseCommandDef(command) {
    var _a, _b, _c;
    var def = command;
    def.argsLookup = {};
    if (def.arguments) {
        def.args = def.arguments.map(parseArgument);
        def.argsLookup = def.args.reduce(function (lu, arg, idx) {
            var _a;
            lu ? (lu[arg.name] = idx) : (lu = (_a = {}, _a[arg.name] = idx, _a));
            return lu;
        }, {});
    }
    if (def.options) {
        def.opts = def.options.reduce(function (opts, opt) {
            var _a;
            // parse option and discard if name matches a positional arg
            var pOpt = parseOption(opt);
            if (((_a = def.argsLookup) === null || _a === void 0 ? void 0 : _a[pOpt.name]) !== undefined) {
                return opts;
            }
            opts ? opts.push(pOpt) : (opts = [pOpt]);
            return opts;
        }, []);
        def.optsLookup = (_a = def.opts) === null || _a === void 0 ? void 0 : _a.reduce(function (lu, opt) {
            if (opt.name) {
                if (!lu)
                    lu = {};
                lu[opt.name] = true;
            }
            if (opt.alias) {
                if (!lu)
                    lu = {};
                lu[opt.alias] = true;
            }
            return lu;
        }, {});
        def.aliasLookup = (_b = def.opts) === null || _b === void 0 ? void 0 : _b.reduce(function (lu, opt) {
            if (opt.alias) {
                if (!lu)
                    lu = {};
                lu[opt.alias] = opt.name;
            }
            return lu;
        }, {});
        (_c = def.opts) === null || _c === void 0 ? void 0 : _c.forEach(function (opt) {
            var _a;
            if (opt.parseOpt.boolean) {
                if (!((_a = def.argsParserOpts) === null || _a === void 0 ? void 0 : _a.boolean)) {
                    def.argsParserOpts = { boolean: [opt.name] };
                }
                else {
                    def.argsParserOpts.boolean.push(opt.name);
                }
            }
        });
        if (!def.argsParserOpts) {
            def.argsParserOpts = {};
        }
        if (def.opts)
            def.argsParserOpts.alias = optionsToArgsParserOpts(def.opts);
    }
    debug_1.debug({ def: def });
    return def;
}
function parseCommandDefs(commands) {
    for (var key in commands) {
        // _forEach(commands, (def, key) => {
        var command = commands[key];
        if (command.commands) {
            parseCommandDefs(command.commands);
        }
        else {
            commands[key] = parseCommandDef(command);
        }
    }
}
function mergeCommands(pluginSet, plugin) {
    var commands = plugin.commands;
    if (commands) {
        parseCommandDefs(commands);
        merge_1.default(pluginSet.commands, plugin.commands);
    }
    return pluginSet;
}
/**
 * Add a plugin to a plugin set
 *
 * @param {Object} pluginSet - registered plugins
 * @param {Object} pluginSet.schema - schema
 * @param {Object} pluginSet.stores - stores
 * @param {Object} pluginSet.commands - commands
 * @param {Object} plugin - plugin to register
 * @param {Object} plugin.schema - schema
 * @param {Object} plugin.stores - stores
 * @param {Object} plugin.commands - commands
 * @returns {Object} the updated pluginSet
 */
function addPlugin(pluginSet, plugin) {
    if (plugin.schema) {
        merge_1.default(pluginSet.schema, plugin.schema);
    }
    if (plugin.stores) {
        merge_1.default(pluginSet.stores, plugin.stores);
    }
    if (plugin.commands) {
        pluginSet = mergeCommands(pluginSet, plugin);
    }
    return pluginSet;
}
/**
 * merge one or more plugins into a plugin set
 *
 * @param {Object[]|Object} plugins - a single plugin or list of plugins
 * @returns {Object} pluginSet
 */
function createPluginSet(plugins) {
    if (!Array.isArray(plugins)) {
        plugins = [plugins];
    }
    return Promise.all(flatten_1.default(plugins)).then(function (pluginList) {
        var pluginSet = { schema: {}, stores: {}, commands: {} };
        pluginList.forEach(function (plugin) {
            pluginSet = addPlugin(pluginSet, plugin);
        });
        return pluginSet;
    });
}
function Makitso(args) {
    var commandPrompt = args.commandPrompt, plugins = args.plugins, cmdLine = args.cmdLine;
    var prompt;
    if (commandPrompt) {
        if (isString_1.default(commandPrompt)) {
            prompt = new makitso_prompt_1.Prompt({ prompt: commandPrompt });
        }
        else {
            prompt = commandPrompt;
        }
    }
    else {
        prompt = new makitso_prompt_1.Prompt();
    }
    return createPluginSet(plugins).then(function (pluginSet) {
        var schema = pluginSet.schema, stores = pluginSet.stores, commands = pluginSet.commands;
        var context = context_1.Context({ schema: schema, stores: stores, commands: commands });
        return command_1.start({ context: context, commands: commands, prompt: prompt, cmdLine: cmdLine });
    });
}
exports.Makitso = Makitso;
